REM
REM  MMU emulation (Memory Management Unit)
REM  This file is part of the ZilogBoy project.
REM  Copyright (C) Mateusz Viste 2010
REM

DIM SHARED MemoryInternalRAM(0 TO 8191) AS UBYTE ' Internal RAM Area  [8KiB]
DIM SHARED MemoryInternalHiRAM(&hFF80 TO &hFFFF) AS UBYTE ' Internal RAM (high area + IE register)
DIM SHARED MemoryBankedRAM(0 TO 2097152) AS UBYTE ' Banked RAM [2MiB]
DIM SHARED MemoryROM(0 TO 2097152) AS UBYTE ' ROM Area  [2MiB]
DIM SHARED VideoRAM(&h8000 TO &h9FFF) AS UBYTE     ' Video RAM [8KiB]
DIM SHARED SpriteOAM(&hFE00 TO &hFE9F) AS UBYTE    ' Sprite OAM memory
DIM SHARED IoRegisters(&hFF00 TO &hFF4B) AS UBYTE  ' All I/O memory-mapped registers
DIM SHARED MemoryMAP(&h0000 TO &hFFFF) AS UBYTE    ' Regular memory (for unmapped regions)

SUB InitRAM()
  REM Init the RAM areas to random values, as in a real GameBoy
  DIM AS INTEGER x
  FOR x = LBOUND(MemoryInternalRAM) TO UBOUND(MemoryInternalRAM)
    MemoryInternalRAM(x) = INT(RND * 256)
  NEXT x
  FOR x = LBOUND(MemoryInternalHiRAM) TO UBOUND(MemoryInternalHiRAM)
    MemoryInternalHiRAM(x) = INT(RND * 256)
  NEXT x
  FOR x = LBOUND(MemoryBankedRAM) TO UBOUND(MemoryBankedRAM)
    MemoryBankedRAM(x) = INT(RND * 256)
  NEXT x
  FOR x = LBOUND(MemoryMAP) TO UBOUND(MemoryMAP)
    MemoryMAP(x) = INT(RND * 256)
  NEXT x
END SUB


FUNCTION MemoryRead(BYREF memaddr AS INTEGER) AS UBYTE
  IF memaddr >= &hC000 AND memaddr < &hE000 THEN        ' Internal 8KiB RAM
    RETURN MemoryInternalRAM(memaddr)
  ELSEIF memaddr >= &hE000 AND memaddr < &hFE00 THEN    ' RAM mirror
    RETURN MemoryInternalRAM(memaddr - 8192)
  ELSEIF memaddr >= &hFE00 AND memaddr < &hFEA0 THEN    ' Sprite OAM memory
    RETURN SpriteOAM(memaddr)
  ELSEIF memaddr >= &hFF80 AND memaddr <= &hFFFF THEN   ' Hi RAM area
    RETURN MemoryInternalHiRAM(memaddr)
  ELSEIF memaddr >= 0 AND memaddr < 16384 THEN          ' ROM bank #0
    RETURN MemoryROM(memaddr)
  ELSEIF memaddr >= &hFF00 AND memaddr <= &hFF4B THEN   ' I/O registers
    RETURN IoRegisters(memaddr)
  ELSEIF memaddr >= 16384 AND memaddr < 32768 THEN      ' ROM bank #n
    IF RomInfos.RomSize = 32768 THEN
        RETURN MemoryROM(memaddr)
      ELSE
        PRINT "Ooops, not supported ROM size! (RomSize = " & RomInfos.RomSize & ")"
        QuitEmulator = 1
    END IF
  ELSEIF memaddr >= &h8000 AND memaddr <= &h9FFF THEN   ' Video RAM (8KiB)
    RETURN VideoRAM(memaddr)
  ELSE                                                  ' Else it's something I don't know about
    'PRINT "MMU fatal error: Tried to read from an unknown memory address: "; HEX(memaddr)
    'QuitEmulator = 1
    RETURN MemoryMAP(memaddr)
  END IF
END FUNCTION


SUB MemoryWrite(BYREF memaddr AS INTEGER, DataByte AS UBYTE)
  DIM AS INTEGER x
  IF memaddr < &h8000 THEN      ' Configuration request for the MBC chip
    REM   As for now, I ignore these. :-/
  ELSEIF memaddr >= &hC000 AND memaddr < &hE000 THEN        ' Internal 8KiB RAM
    MemoryInternalRAM(memaddr) = DataByte
  ELSEIF memaddr >= &hE000 AND memaddr < &hFE00 THEN    ' RAM mirror
    MemoryInternalRAM(memaddr - 8192) = DataByte
  ELSEIF memaddr >= &hFE00 AND memaddr < &hFEA0 THEN    ' Sprite OAM memory
    SpriteOAM(memaddr) = DataByte
  ELSEIF memaddr >= &hFF80 AND memaddr <= &hFFFF THEN   ' Hi RAM area
    MemoryInternalHiRAM(memaddr) = DataByte
  ELSEIF memaddr = &hFF44 THEN   ' CURLINE [RW] Current Scanline.
    IoRegisters(memaddr) = 0     ' Writing into this register resets it.
  ELSEIF memaddr = &hFF46 THEN ' Starts LCD OAM DMA transfer
    IF DataByte <= &hF1 THEN   ' Normally, should take 160us... but who cares?
        FOR x = 0 TO 159       ' Let's copy XX00-XX9F to FE00-FE9F
          SpriteOAM(&hFE00 + x) = MemoryRead((DataByte * 256) + x)
        NEXT x
      ELSE
        PRINT "MMU FATAL ERROR! Tried to run LCD OAM DMA transfer using an invalid range!"
        SLEEP
        END
    END IF
  ELSEIF memaddr >= &hFF00 AND memaddr <= &hFF4B THEN   ' I/O registers
    IoRegisters(memaddr) = DataByte
  ELSEIF memaddr >= &h8000 AND memaddr <= &h9FFF THEN   ' Video RAM (8KiB)
    VideoRAM(memaddr) = DataByte
  ELSE                                                  ' Else it's something I don't know about
    'PRINT "MMU fatal error: Tried to write to an unknown memory address: "; HEX(memaddr)
    'QuitEmulator = 1
    MemoryMAP(memaddr) = DataByte
  END IF
END SUB

